import os
import sys
import cv2
import numpy as np
import tkinter as tk
from tkinter import filedialog, messagebox, ttk
from PIL import Image, ImageTk
import threading
import time
import keyboard
from datetime import timedelta


class MultiCamSwitcher:
    def __init__(self, root):
        self.root = root
        self.root.title("MultiCam Switcher")
        self.root.geometry("1280x800")
        self.root.configure(bg="#2b2b2b")  # Dark theme background

        # Initialize variables
        self.tracks = [None] * 22
        self.current_track = 0
        self.is_recording = False
        self.is_paused = False
        self.is_playing = False
        self.output_frames = []
        self.playback_thread = None
        self.recording_thread = None

        # Playback variables
        self.current_frame = 0
        self.max_frames = 0
        self.sync_position = 0  # current position in milliseconds
        self.global_fps = 30  # default FPS, will update based on clips

        # Key mappings
        self.key_mappings = {
            "f1": 0, "f2": 1, "f3": 2, "f4": 3, "f5": 4, "f6": 5,
            "f7": 6, "f8": 7, "f9": 8, "f10": 9, "f11": 10, "f12": 11,
            "num_0": 12, "num_1": 13, "num_2": 14, "num_3": 15, "num_4": 16,
            "num_5": 17, "num_6": 18, "num_7": 19, "num_8": 20, "num_9": 21
        }

        # Frame dimensions
        self.preview_width = 640
        self.preview_height = 360

        # Create UI elements
        self.create_ui()
        # Set up keyboard listeners
        self.setup_keyboard_listeners()
        # Set up mouse wheel zoom (and timeline seeking)
        self.setup_mouse_zoom()

    def create_ui(self):
        self.create_menu()

        main_container = tk.Frame(self.root, bg="#2b2b2b")
        main_container.pack(fill=tk.BOTH, expand=True, padx=5, pady=5)

        # Top area: Preview and controls
        top_frame = tk.Frame(main_container, bg="#2b2b2b")
        top_frame.pack(fill=tk.BOTH, expand=True)

        # Preview panel
        preview_frame = tk.LabelFrame(top_frame, text="Preview", fg="white", bg="#333333", bd=2)
        preview_frame.pack(side=tk.LEFT, fill=tk.BOTH, expand=True, padx=5, pady=5)
        self.preview_canvas = tk.Canvas(preview_frame, width=self.preview_width, height=self.preview_height,
                                        bg="black", highlightthickness=0)
        self.preview_canvas.pack(padx=10, pady=10)

        # Controls panel
        controls_frame = tk.LabelFrame(top_frame, text="Controls", fg="white", bg="#333333", bd=2)
        controls_frame.pack(side=tk.RIGHT, fill=tk.Y, padx=5, pady=5)

        # Info frame: current track and timecode display
        info_frame = tk.Frame(controls_frame, bg="#333333")
        info_frame.pack(fill=tk.X, padx=10, pady=10)
        self.current_track_label = tk.Label(info_frame, text="Current Track: None",
                                            bg="#333333", fg="white", font=("Arial", 10, "bold"))
        self.current_track_label.pack(anchor=tk.W)
        self.timecode_label = tk.Label(info_frame, text="00:00:00.000",
                                       bg="#333333", fg="#00ff00", font=("Courier", 14, "bold"))
        self.timecode_label.pack(anchor=tk.W, pady=5)

        # Transport controls
        transport_frame = tk.Frame(controls_frame, bg="#333333")
        transport_frame.pack(fill=tk.X, padx=10, pady=10)
        self.play_btn = tk.Button(transport_frame, text="▶ Play (Space)",
                                  command=self.toggle_play, width=10,
                                  bg="#444444", fg="white", activebackground="#666666")
        self.play_btn.pack(side=tk.LEFT, padx=2)
        stop_playback_btn = tk.Button(transport_frame, text="■ Stop",
                                      command=self.stop_playback, width=10,
                                      bg="#444444", fg="white", activebackground="#666666")
        stop_playback_btn.pack(side=tk.LEFT, padx=2)

        # Recording controls
        recording_frame = tk.Frame(controls_frame, bg="#333333")
        recording_frame.pack(fill=tk.X, padx=10, pady=10)
        self.record_btn = tk.Button(recording_frame, text="⚫ Record",
                                    command=self.toggle_recording, width=10,
                                    bg="#8b0000", fg="white", activebackground="#ff0000")
        self.record_btn.pack(side=tk.LEFT, padx=2)
        self.pause_btn = tk.Button(recording_frame, text="⏸ Pause",
                                   command=self.toggle_pause, width=10,
                                   bg="#444444", fg="white", activebackground="#666666", state=tk.DISABLED)
        self.pause_btn.pack(side=tk.LEFT, padx=2)
        self.stop_btn = tk.Button(recording_frame, text="⏹ Stop",
                                  command=self.stop_recording, width=10,
                                  bg="#444444", fg="white", activebackground="#666666", state=tk.DISABLED)
        self.stop_btn.pack(side=tk.LEFT, padx=2)

        # Bottom area: Timeline and tracks
        bottom_frame = tk.Frame(main_container, bg="#2b2b2b")
        bottom_frame.pack(fill=tk.BOTH, expand=True, pady=5)
        timeline_frame = tk.LabelFrame(bottom_frame, text="Timeline", fg="white",
                                       bg="#333333", bd=2)
        timeline_frame.pack(fill=tk.BOTH, expand=True, padx=5, pady=5)

        timeline_container = tk.Frame(timeline_frame, bg="#333333")
        timeline_container.pack(fill=tk.BOTH, expand=True)
        y_scrollbar = tk.Scrollbar(timeline_container, orient=tk.VERTICAL)
        y_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        x_scrollbar = tk.Scrollbar(timeline_container, orient=tk.HORIZONTAL)
        x_scrollbar.pack(side=tk.BOTTOM, fill=tk.X)
        self.timeline_canvas = tk.Canvas(
            timeline_container, bg="#1e1e1e",
            yscrollcommand=y_scrollbar.set, xscrollcommand=x_scrollbar.set
        )
        self.timeline_canvas.pack(fill=tk.BOTH, expand=True)
        y_scrollbar.config(command=self.timeline_canvas.yview)
        x_scrollbar.config(command=self.timeline_canvas.xview)

        self.timeline_content = tk.Frame(self.timeline_canvas, bg="#1e1e1e")
        self.timeline_canvas.create_window((0, 0), window=self.timeline_content, anchor=tk.NW)

        # Time ruler
        self.ruler_canvas = tk.Canvas(self.timeline_content, height=20, bg="#2d2d2d", highlightthickness=0)
        self.ruler_canvas.pack(fill=tk.X)

        # Create 22 timeline tracks
        self.track_frames = []
        self.track_canvases = []
        for i in range(22):
            track_row = tk.Frame(self.timeline_content, bg="#1e1e1e")
            track_row.pack(fill=tk.X)
            key_label = f"F{i+1}" if i < 12 else f"Num {i-12}"
            track_label = tk.Label(track_row, text=f"Track {i+1}\n({key_label})",
                                   width=10, height=2, bg="#444444", fg="white")
            track_label.pack(side=tk.LEFT)
            track_canvas = tk.Canvas(track_row, height=40, bg="#2d2d2d", highlightthickness=0)
            track_canvas.pack(side=tk.LEFT, fill=tk.X, expand=True, padx=(0, 5), pady=2)
            add_btn = tk.Button(track_row, text="+", command=lambda i=i: self.add_clip_to_track(i),
                                width=3, bg="#555555", fg="white")
            track_canvas.create_window(10, 20, window=add_btn)
            self.track_frames.append(track_row)
            self.track_canvases.append(track_canvas)

        self.timeline_content.update_idletasks()
        self.timeline_canvas.config(scrollregion=self.timeline_canvas.bbox(tk.ALL))
        self.playhead = self.ruler_canvas.create_line(0, 0, 0, 1000, fill="#ff0000", width=2)
        self.status_bar = tk.Label(self.root, text="Ready", bd=1, relief=tk.SUNKEN,
                                   anchor=tk.W, bg="#333333", fg="white")
        self.status_bar.pack(side=tk.BOTTOM, fill=tk.X)
        self.zoom_indicator = tk.Label(self.status_bar, text="Zoom: 100%",
                                       bg="#333333", fg="#aaaaaa", padx=10)
        self.zoom_indicator.pack(side=tk.RIGHT)
        self.timeline_scale = 10  # pixels per frame
        self.update_timeline_scale()

    def create_menu(self):
        menu_bar = tk.Menu(self.root)
        file_menu = tk.Menu(menu_bar, tearoff=0)
        file_menu.add_command(label="New Project", command=self.new_project)
        file_menu.add_separator()
        file_menu.add_command(label="Exit", command=self.root.quit)
        menu_bar.add_cascade(label="File", menu=file_menu)

        timeline_menu = tk.Menu(menu_bar, tearoff=0)
        timeline_menu.add_command(label="Zoom In (Ctrl+Mouse Wheel Up)", command=lambda: self.zoom_timeline(1.2))
        timeline_menu.add_command(label="Zoom Out (Ctrl+Mouse Wheel Down)", command=lambda: self.zoom_timeline(0.8))
        timeline_menu.add_command(label="Reset Zoom", command=lambda: self.zoom_timeline(1, reset=True))
        menu_bar.add_cascade(label="Timeline", menu=timeline_menu)
        self.root.config(menu=menu_bar)

    def setup_keyboard_listeners(self):
        for key, track_idx in self.key_mappings.items():
            keyboard.on_press_key(key, lambda e, track_idx=track_idx: self.switch_to_track(track_idx))

    def setup_mouse_zoom(self):
        # Bind mouse wheel events for zooming and timeline seeking for multiple canvases
        self.timeline_canvas.bind("<Control-MouseWheel>", self.handle_mouse_zoom)
        self.ruler_canvas.bind("<Control-MouseWheel>", self.handle_mouse_zoom)
        self.timeline_canvas.bind("<Control-Button-4>", lambda event: self.handle_mouse_zoom(event, direction=1))
        self.timeline_canvas.bind("<Control-Button-5>", lambda event: self.handle_mouse_zoom(event, direction=-1))
        self.ruler_canvas.bind("<Control-Button-4>", lambda event: self.handle_mouse_zoom(event, direction=1))
        self.ruler_canvas.bind("<Control-Button-5>", lambda event: self.handle_mouse_zoom(event, direction=-1))
        for canvas in self.track_canvases:
            canvas.bind("<Control-MouseWheel>", self.handle_mouse_zoom)
            canvas.bind("<Control-Button-4>", lambda event: self.handle_mouse_zoom(event, direction=1))
            canvas.bind("<Control-Button-5>", lambda event: self.handle_mouse_zoom(event, direction=-1))
        self.ruler_canvas.bind("<Button-1>", self.seek_timeline)
        for canvas in self.track_canvases:
            canvas.bind("<Button-1>", self.seek_timeline)
        self.root.bind("<space>", self.toggle_play_space)

    def handle_mouse_zoom(self, event, direction=None):
        if direction is None:
            direction = 1 if event.delta > 0 else -1
        zoom_factor = 1.1 if direction > 0 else 0.9
        self.zoom_timeline(zoom_factor)
        return "break"

    def new_project(self):
        if messagebox.askyesno("New Project", "Start a new project? Any unsaved changes will be lost."):
            self.cleanup()
            self.tracks = [None] * 22
            self.current_track = 0
            self.is_recording = False
            self.is_paused = False
            self.is_playing = False
            self.output_frames = []
            self.current_frame = 0
            self.max_frames = 0
            self.sync_position = 0
            self.update_timeline()
            self.update_status("New project created")

    def add_clip_to_track(self, track_idx):
        file_path = filedialog.askopenfilename(
            title=f"Select video for Track {track_idx+1}",
            filetypes=[("Video files", "*.mp4 *.avi *.mov *.mkv")]
        )
        if file_path:
            try:
                cap = cv2.VideoCapture(file_path)
                if not cap.isOpened():
                    messagebox.showerror("Error", f"Cannot open video file: {file_path}")
                    return
                fps = cap.get(cv2.CAP_PROP_FPS) or 30
                width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
                height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
                frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                duration_ms = (frame_count / fps) * 1000
                if track_idx == 0 or self.tracks[0] is None:
                    self.global_fps = fps
                self.tracks[track_idx] = {
                    "path": file_path,
                    "fps": fps,
                    "width": width,
                    "height": height,
                    "frame_count": frame_count,
                    "duration_ms": duration_ms,
                    "cap": cap,
                    "current_ms": 0
                }
                self.update_max_frames()
                self.draw_clip_on_timeline(track_idx)
                filename = os.path.basename(file_path)
                self.update_status(f"Added {filename} to Track {track_idx+1}")
            except Exception as e:
                messagebox.showerror("Error", f"Error loading video: {str(e)}")

    def update_max_frames(self):
        max_duration_ms = 0
        for track in self.tracks:
            if track is not None:
                max_duration_ms = max(max_duration_ms, track["duration_ms"])
        self.max_frames = int((max_duration_ms / 1000) * self.global_fps)
        self.update_timeline_scale()

    def draw_clip_on_timeline(self, track_idx):
        if self.tracks[track_idx] is None:
            return
        track_canvas = self.track_canvases[track_idx]
        track_canvas.delete("clip")
        clip = self.tracks[track_idx]
        duration_frames = int((clip["duration_ms"] / 1000) * self.global_fps)
        clip_width = duration_frames * self.timeline_scale
        filename = os.path.basename(clip["path"])
        if len(filename) > 20:
            filename = filename[:17] + "..."
        track_canvas.create_rectangle(10, 5, 10 + clip_width, 35,
                                      fill="#4b6584", outline="#dfe6e9", tags="clip")
        track_canvas.create_text(20, 20, text=filename, anchor=tk.W,
                                   fill="white", tags="clip")
        self.timeline_content.update_idletasks()
        self.timeline_canvas.config(scrollregion=self.timeline_canvas.bbox(tk.ALL))

    def update_timeline_scale(self):
        self.ruler_canvas.delete("all")
        total_width = self.max_frames * self.timeline_scale
        self.ruler_canvas.config(width=max(total_width, 800))
        marker_interval_frames = max(int(self.global_fps * 5), 1)
        for frame in range(0, self.max_frames + 1, marker_interval_frames):
            x_pos = frame * self.timeline_scale
            self.ruler_canvas.create_line(x_pos, 15, x_pos, 20, fill="white")
            seconds = frame / self.global_fps
            time_str = str(timedelta(seconds=seconds)).split('.')[0]
            self.ruler_canvas.create_text(x_pos, 10, text=time_str, fill="white", font=("Arial", 8))
        for i, track_canvas in enumerate(self.track_canvases):
            track_canvas.config(width=max(total_width, 800))
            if self.tracks[i] is not None:
                self.draw_clip_on_timeline(i)
        self.timeline_content.update_idletasks()
        self.timeline_canvas.config(scrollregion=self.timeline_canvas.bbox(tk.ALL))
        base_scale = 10  # default value
        zoom_percentage = int((self.timeline_scale / base_scale) * 100)
        self.zoom_indicator.config(text=f"Zoom: {zoom_percentage}%")

    def zoom_timeline(self, factor, reset=False):
        if reset:
            self.timeline_scale = 10
        else:
            visible_left = self.timeline_canvas.canvasx(0)
            visible_width = self.timeline_canvas.winfo_width()
            center_x = visible_left + visible_width / 2
            old_scale = self.timeline_scale
            self.timeline_scale = max(5, min(50, self.timeline_scale * factor))
            scale_factor = self.timeline_scale / old_scale
            new_center_x = center_x * scale_factor
        self.update_timeline_scale()
        self.timeline_canvas.update_idletasks()
        new_visible_width = self.timeline_canvas.winfo_width()
        self.timeline_canvas.xview_moveto((max(0, new_center_x - new_visible_width / 2)) / self.ruler_canvas.winfo_width())

    def update_timeline(self):
        for i in range(22):
            if self.tracks[i] is not None:
                self.draw_clip_on_timeline(i)
            else:
                self.track_canvases[i].delete("clip")
        self.update_timeline_scale()

    def update_playhead(self):
        if self.global_fps > 0:
            current_frame = int((self.sync_position / 1000) * self.global_fps)
            x_position = current_frame * self.timeline_scale
            self.ruler_canvas.coords(self.playhead, x_position, 0, x_position, 1000)
            self.timeline_canvas.xview_moveto(max(0, (x_position - 400) / self.ruler_canvas.winfo_width()))
        ms = self.sync_position
        hours = int(ms / 3600000)
        ms %= 3600000
        minutes = int(ms / 60000)
        ms %= 60000
        seconds = int(ms / 1000)
        ms %= 1000
        timecode = f"{hours:02d}:{minutes:02d}:{seconds:02d}.{int(ms):03d}"
        self.timecode_label.config(text=timecode)

    def switch_to_track(self, track_idx):
        if 0 <= track_idx < 22 and self.tracks[track_idx] is not None:
            self.current_track = track_idx
            self.current_track_label.config(text=f"Current Track: {track_idx+1}")
            if self.is_recording and not self.is_paused:
                self.output_frames.append(("switch", track_idx, self.sync_position))
            self.update_status(f"Switched to Track {track_idx+1}")

    def toggle_play(self):
        if not self.is_playing:
            if not any(track is not None for track in self.tracks):
                messagebox.showerror("Error", "No tracks loaded. Please add video clips first.")
                return
            if self.max_frames > 0 and (self.sync_position / 1000 * self.global_fps) >= self.max_frames:
                self.sync_position = 0
            for track in self.tracks:
                if track is not None:
                    track["cap"].set(cv2.CAP_PROP_POS_FRAMES, 0)
                    track["current_ms"] = 0
            self.is_playing = True
            self.play_btn.config(text="⏸ Pause")
            if self.playback_thread is None or not self.playback_thread.is_alive():
                self.playback_thread = threading.Thread(target=self.playback_loop)
                self.playback_thread.daemon = True
                self.playback_thread.start()
        else:
            self.is_playing = False
            self.play_btn.config(text="▶ Play")

    def stop_playback(self):
        self.is_playing = False
        self.play_btn.config(text="▶ Play")
        self.sync_position = 0
        for track in self.tracks:
            if track is not None:
                track["current_ms"] = 0
                track["cap"].set(cv2.CAP_PROP_POS_FRAMES, 0)
        self.update_playhead()
        self.update_status("Playback stopped")

    def playback_loop(self):
        last_update_time = time.time()
        while self.is_playing:
            if not self.is_paused or not self.is_recording:
                current_time = time.time()
                elapsed_ms = (current_time - last_update_time) * 1000
                last_update_time = current_time
                self.sync_position += elapsed_ms
                if self.max_frames > 0 and (self.sync_position / 1000 * self.global_fps) >= self.max_frames:
                    # If recording is active, stop it before stopping playback
                    if self.is_recording:
                        self.is_recording = False
                        self.is_paused = False
                        self.record_btn.config(text="⚫ Record", bg="#8b0000")
                        self.pause_btn.config(text="⏸ Pause", state=tk.DISABLED)
                        self.stop_btn.config(state=tk.DISABLED)
                        if self.output_frames:
                            self.update_status("Recording stopped. Preparing to save...")
                            output_path = filedialog.asksaveasfilename(
                                title="Save Recorded Video",
                                defaultextension=".mp4",
                                filetypes=[("MP4 video", "*.mp4")]
                            )
                            if output_path:
                                save_thread = threading.Thread(target=self.save_recording, args=(output_path,))
                                save_thread.daemon = True
                                save_thread.start()
                            else:
                                self.update_status("Recording canceled")
                    self.stop_playback()
                    break
                if self.tracks[self.current_track] is not None:
                    track = self.tracks[self.current_track]
                    desired_frame = int((self.sync_position / 1000) * track["fps"])
                    if desired_frame >= track["frame_count"]:
                        blank_frame = np.zeros((self.preview_height, self.preview_width, 3), dtype=np.uint8)
                        self.update_preview(blank_frame)
                    else:
                        track["cap"].set(cv2.CAP_PROP_POS_FRAMES, desired_frame)
                        ret, frame = track["cap"].read()
                        if ret:
                            track["current_ms"] = self.sync_position
                            frame = cv2.resize(frame, (self.preview_width, self.preview_height))
                            self.update_preview(frame)
                            if self.is_recording and not self.is_paused:
                                self.output_frames.append(("frame", self.current_track, frame.copy()))
                    self.update_playhead()
                    time.sleep(1 / self.global_fps)
                else:
                    last_update_time = time.time()
                    time.sleep(0.1)
            else:
                last_update_time = time.time()
                time.sleep(0.1)

    def update_preview(self, frame):
        # Convert BGR frame to RGB for Tkinter display
        frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        img = Image.fromarray(frame_rgb)
        img_tk = ImageTk.PhotoImage(image=img)
        self.preview_canvas.create_image(0, 0, anchor=tk.NW, image=img_tk)
        self.preview_canvas.image = img_tk  # keep a reference

    def toggle_recording(self):
        if not self.is_recording:
            if not any(track is not None for track in self.tracks):
                messagebox.showerror("Error", "No tracks loaded. Please add video clips first.")
                return
            if self.tracks[0] is None:
                messagebox.showerror("Error", "Track 1 must have a clip for recording (needed for output settings).")
                return
            if not self.is_playing:
                self.toggle_play()
            self.output_frames = []
            self.is_recording = True
            self.is_paused = False
            self.output_frames.append(("switch", self.current_track, self.sync_position))
            self.record_btn.config(text="⚫ Recording", bg="#ff0000")
            self.pause_btn.config(state=tk.NORMAL)
            self.stop_btn.config(state=tk.NORMAL)
            self.update_status("Recording started")
        else:
            self.stop_recording()

    def toggle_pause(self):
        if self.is_recording:
            self.is_paused = not self.is_paused
            if self.is_paused:
                self.pause_btn.config(text="▶ Resume")
                self.update_status("Recording paused")
                if self.is_playing:
                    self.toggle_play()
            else:
                self.pause_btn.config(text="⏸ Pause")
                self.update_status("Recording resumed")
                if not self.is_playing:
                    self.toggle_play()

    def stop_recording(self):
        if self.is_recording:
            self.is_recording = False
            self.is_paused = False
            self.record_btn.config(text="⚫ Record", bg="#8b0000")
            self.pause_btn.config(text="⏸ Pause", state=tk.DISABLED)
            self.stop_btn.config(state=tk.DISABLED)
            if self.output_frames:
                self.update_status("Recording stopped. Preparing to save...")
                output_path = filedialog.asksaveasfilename(
                    title="Save Recorded Video",
                    defaultextension=".mp4",
                    filetypes=[("MP4 video", "*.mp4")]
                )
                if output_path:
                    save_thread = threading.Thread(target=self.save_recording, args=(output_path,))
                    save_thread.daemon = True
                    save_thread.start()
                else:
                    self.update_status("Recording canceled")
            else:
                self.update_status("Recording stopped (no frames recorded)")

    def save_recording(self, output_path):
        if not self.output_frames:
            messagebox.showinfo("Information", "No frames were recorded.")
            return
        try:
            ref_track = self.tracks[0]
            width = ref_track["width"]
            height = ref_track["height"]
            fps = ref_track["fps"]
            # Use a fixed codec 'mp4v', which is known to be compatible on Windows.
            fourcc = cv2.VideoWriter_fourcc(*"mp4v")
            output_file = os.path.splitext(output_path)[0] + ".mp4"
            out = cv2.VideoWriter(output_file, fourcc, fps, (width, height), True)
            if not out.isOpened():
                raise Exception("VideoWriter failed to open with mp4v codec. Check your system codecs.")
            for action, _, frame in self.output_frames:
                if action == "frame":
                    out.write(frame)
            out.release()
            self.update_status(f"Recording saved successfully to {output_file}")
            messagebox.showinfo("Recording Saved", f"Recording saved successfully to {output_file}")
            self.output_frames.clear()
        except Exception as e:
            messagebox.showerror("Error", f"Failed to save recording: {str(e)}")
            self.update_status("Error saving recording")

    def update_status(self, message):
        zoom_percentage = int((self.timeline_scale / 10) * 100)
        self.zoom_indicator.config(text=f"Zoom: {zoom_percentage}%")
        self.status_bar.config(text=message)
        self.root.update_idletasks()

    def cleanup(self):
        self.is_playing = False
        self.is_recording = False
        for track in self.tracks:
            if track is not None and "cap" in track:
                track["cap"].release()

    def seek_timeline(self, event):
        canvas = event.widget
        click_x = canvas.canvasx(event.x)
        frame_position = int(click_x / self.timeline_scale)
        self.sync_position = (frame_position / self.global_fps) * 1000
        for track in self.tracks:
            if track is not None:
                desired_frame = int((self.sync_position / 1000) * track["fps"])
                track["cap"].set(cv2.CAP_PROP_POS_FRAMES, desired_frame)
                track["current_ms"] = self.sync_position
        self.update_playhead()
        if self.is_recording:
            return
        if self.tracks[self.current_track] is not None:
            track = self.tracks[self.current_track]
            ret, frame = track["cap"].read()
            if ret:
                frame = cv2.resize(frame, (self.preview_width, self.preview_height))
                self.update_preview(frame)

    def toggle_play_space(self, event=None):
        if isinstance(self.root.focus_get(), (tk.Entry, tk.Text)):
            return
        self.toggle_play()


if __name__ == "__main__":
    try:
        import cv2
        import numpy as np
        from PIL import Image, ImageTk
        import keyboard
    except ImportError as e:
        print(f"Missing dependency: {str(e)}")
        print("Please install required dependencies with:")
        print("python -m pip install opencv-python pillow numpy keyboard")
        sys.exit(1)
    root = tk.Tk()
    app = MultiCamSwitcher(root)

    def on_closing():
        app.cleanup()
        root.destroy()

    root.protocol("WM_DELETE_WINDOW", on_closing)
    root.mainloop()
