import os
import sys
import cv2
import numpy as np
import tkinter as tk
from tkinter import filedialog, messagebox
from PIL import Image, ImageTk
import threading
import time
import keyboard

class MultiCamSwitcher:
    def __init__(self, root):
        self.root = root
        self.root.title("MultiCam Switcher")
        self.root.geometry("1200x800")

        # Initialize variables
        self.tracks = [None] * 22  # 22 tracks capability
        self.current_track = 0
        self.is_recording = False
        self.is_paused = False
        self.output_frames = []
        self.preview_thread = None
        self.recording_thread = None
        
        # Key mappings
        self.key_mappings = {
            'f1': 0, 'f2': 1, 'f3': 2, 'f4': 3, 'f5': 4, 'f6': 5, 
            'f7': 6, 'f8': 7, 'f9': 8, 'f10': 9, 'f11': 10, 'f12': 11,
            'num_0': 12, 'num_1': 13, 'num_2': 14, 'num_3': 15, 'num_4': 16,
            'num_5': 17, 'num_6': 18, 'num_7': 19, 'num_8': 20, 'num_9': 21
        }
        
        # Frame width and height
        self.frame_width = 640
        self.frame_height = 360
        
        # Create UI elements
        self.create_ui()
        
        # Set up keyboard listeners
        self.setup_keyboard_listeners()

    def create_ui(self):
        # Main frame
        main_frame = tk.Frame(self.root)
        main_frame.pack(fill=tk.BOTH, expand=True)
        
        # Left panel for tracks
        tracks_frame = tk.LabelFrame(main_frame, text="Tracks")
        tracks_frame.pack(side=tk.LEFT, fill=tk.BOTH, expand=True, padx=5, pady=5)
        
        # Create track buttons
        self.track_buttons = []
        for i in range(22):
            if i < 12:
                key_label = f"F{i+1}"
            else:
                key_label = f"Num {i-12}"
                
            track_frame = tk.Frame(tracks_frame)
            track_frame.pack(fill=tk.X, padx=2, pady=2)
            
            label = tk.Label(track_frame, text=f"Track {i+1} ({key_label}):")
            label.pack(side=tk.LEFT)
            
            btn = tk.Button(track_frame, text="Add Clip", command=lambda i=i: self.add_clip_to_track(i))
            btn.pack(side=tk.LEFT, padx=5)
            
            status = tk.Label(track_frame, text="No clip", width=30, anchor='w')
            status.pack(side=tk.LEFT, padx=5)
            
            self.track_buttons.append((btn, status))
        
        # Right panel for preview and controls
        right_panel = tk.Frame(main_frame)
        right_panel.pack(side=tk.RIGHT, fill=tk.BOTH, expand=True, padx=5, pady=5)
        
        # Preview area
        preview_frame = tk.LabelFrame(right_panel, text="Preview")
        preview_frame.pack(fill=tk.BOTH, expand=True, padx=5, pady=5)
        
        self.preview_canvas = tk.Canvas(preview_frame, width=self.frame_width, height=self.frame_height, bg="black")
        self.preview_canvas.pack(padx=5, pady=5)
        
        # Current track label
        self.current_track_label = tk.Label(preview_frame, text="Current Track: None")
        self.current_track_label.pack(pady=5)
        
        # Controls frame
        controls_frame = tk.LabelFrame(right_panel, text="Controls")
        controls_frame.pack(fill=tk.X, padx=5, pady=5)
        
        # Recording controls
        self.record_btn = tk.Button(controls_frame, text="Record", command=self.toggle_recording)
        self.record_btn.pack(side=tk.LEFT, padx=5, pady=5)
        
        self.pause_btn = tk.Button(controls_frame, text="Pause", command=self.toggle_pause, state=tk.DISABLED)
        self.pause_btn.pack(side=tk.LEFT, padx=5, pady=5)
        
        self.stop_btn = tk.Button(controls_frame, text="Stop", command=self.stop_recording, state=tk.DISABLED)
        self.stop_btn.pack(side=tk.LEFT, padx=5, pady=5)
        
        # Status bar
        self.status_bar = tk.Label(self.root, text="Ready", bd=1, relief=tk.SUNKEN, anchor=tk.W)
        self.status_bar.pack(side=tk.BOTTOM, fill=tk.X)

    def setup_keyboard_listeners(self):
        for key, track_idx in self.key_mappings.items():
            keyboard.on_press_key(key, lambda e, track_idx=track_idx: self.switch_to_track(track_idx))

    def add_clip_to_track(self, track_idx):
        file_path = filedialog.askopenfilename(
            title=f"Select video for Track {track_idx+1}",
            filetypes=[("Video files", "*.mp4 *.avi *.mov *.mkv")]
        )
        
        if file_path:
            try:
                cap = cv2.VideoCapture(file_path)
                if not cap.isOpened():
                    messagebox.showerror("Error", f"Cannot open video file: {file_path}")
                    return
                
                fps = cap.get(cv2.CAP_PROP_FPS)
                width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
                height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
                frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                
                self.tracks[track_idx] = {
                    "path": file_path,
                    "fps": fps,
                    "width": width,
                    "height": height,
                    "frame_count": frame_count,
                    "cap": cap
                }
                
                filename = os.path.basename(file_path)
                if len(filename) > 20:
                    filename = filename[:17] + "..."
                    
                self.track_buttons[track_idx][1].config(text=filename)
                self.update_status(f"Added {os.path.basename(file_path)} to Track {track_idx+1}")
                
                # If this is the first track, start the preview
                if track_idx == 0 and self.preview_thread is None:
                    self.preview_thread = threading.Thread(target=self.preview_loop)
                    self.preview_thread.daemon = True
                    self.preview_thread.start()
            except Exception as e:
                messagebox.showerror("Error", f"Error loading video: {str(e)}")

    def switch_to_track(self, track_idx):
        if 0 <= track_idx < 22 and self.tracks[track_idx] is not None:
            self.current_track = track_idx
            self.current_track_label.config(text=f"Current Track: {track_idx+1}")
            self.update_status(f"Switched to Track {track_idx+1}")

    def preview_loop(self):
        while True:
            if not self.is_paused or not self.is_recording:
                if self.tracks[self.current_track] is not None:
                    cap = self.tracks[self.current_track]["cap"]
                    
                    # Reset to start if at end
                    if cap.get(cv2.CAP_PROP_POS_FRAMES) >= cap.get(cv2.CAP_PROP_FRAME_COUNT) - 1:
                        cap.set(cv2.CAP_PROP_POS_FRAMES, 0)
                    
                    ret, frame = cap.read()
                    if ret:
                        # Resize frame for preview
                        frame = cv2.resize(frame, (self.frame_width, self.frame_height))
                        
                        # Convert frame for display
                        frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                        img = Image.fromarray(frame_rgb)
                        img_tk = ImageTk.PhotoImage(image=img)
                        
                        # Update canvas with new frame
                        self.preview_canvas.create_image(0, 0, anchor=tk.NW, image=img_tk)
                        self.preview_canvas.image = img_tk  # Keep reference
                        
                        # If recording, add frame to output
                        if self.is_recording and not self.is_paused:
                            self.output_frames.append((self.current_track, frame))
                    
                    # Control frame rate
                    fps = self.tracks[self.current_track]["fps"]
                    time.sleep(1/fps)
                else:
                    time.sleep(0.1)  # Sleep if no track is active
            else:
                time.sleep(0.1)  # Sleep when paused

    def toggle_recording(self):
        if not self.is_recording:
            # Check if there are any tracks loaded
            if not any(track is not None for track in self.tracks):
                messagebox.showerror("Error", "No tracks loaded. Please add at least one video clip.")
                return
                
            # Check if track 0 has a clip
            if self.tracks[0] is None:
                messagebox.showerror("Error", "Track 1 must have a clip for recording.")
                return
                
            self.output_frames = []
            self.is_recording = True
            self.is_paused = False
            self.record_btn.config(text="Recording...", state=tk.DISABLED)
            self.pause_btn.config(state=tk.NORMAL)
            self.stop_btn.config(state=tk.NORMAL)
            
            # Reset all videos to start
            for track in self.tracks:
                if track is not None:
                    track["cap"].set(cv2.CAP_PROP_POS_FRAMES, 0)
            
            self.update_status("Recording started")
            
            # Start recording thread
            if self.recording_thread is None:
                self.recording_thread = threading.Thread(target=self.recording_loop)
                self.recording_thread.daemon = True
                self.recording_thread.start()

    def toggle_pause(self):
        if self.is_recording:
            self.is_paused = not self.is_paused
            if self.is_paused:
                self.pause_btn.config(text="Resume")
                self.update_status("Recording paused")
            else:
                self.pause_btn.config(text="Pause")
                self.update_status("Recording resumed")

    def recording_loop(self):
        while self.is_recording:
            time.sleep(0.1)  # Check status periodically

    def stop_recording(self):
        if self.is_recording:
            self.is_recording = False
            self.is_paused = False
            self.record_btn.config(text="Record", state=tk.NORMAL)
            self.pause_btn.config(text="Pause", state=tk.DISABLED)
            self.stop_btn.config(state=tk.DISABLED)
            
            self.update_status("Recording stopped. Preparing to save...")
            
            # Get save location
            output_path = filedialog.asksaveasfilename(
                title="Save Recorded Video",
                defaultextension=".mp4",
                filetypes=[("MP4 video", "*.mp4")]
            )
            
            if output_path:
                self.save_recording(output_path)
            
            # Reset recording thread
            self.recording_thread = None

    def save_recording(self, output_path):
        if not self.output_frames:
            messagebox.showinfo("Information", "No frames were recorded.")
            return
            
        try:
            # Use properties from the first track for output
            ref_track = self.tracks[0]
            width = ref_track["width"]
            height = ref_track["height"]
            fps = ref_track["fps"]
            
            # Create video writer
            fourcc = cv2.VideoWriter_fourcc(*'mp4v')
            out = cv2.VideoWriter(output_path, fourcc, fps, (width, height))
            
            # Status variables for progress
            total_frames = len(self.output_frames)
            progress_step = max(1, total_frames // 100)
            
            # Write frames to output
            for i, (track_idx, frame) in enumerate(self.output_frames):
                # Resize frame to match output dimensions
                if frame.shape[1] != width or frame.shape[0] != height:
                    frame = cv2.resize(frame, (width, height))
                    
                out.write(frame)
                
                # Update status periodically
                if i % progress_step == 0:
                    progress = (i / total_frames) * 100
                    self.update_status(f"Saving recording: {progress:.1f}%")
                
            # Release writer
            out.release()
            self.update_status(f"Recording saved to {output_path}")
            messagebox.showinfo("Success", f"Recording saved to {output_path}")
            
        except Exception as e:
            messagebox.showerror("Error", f"Error saving video: {str(e)}")
            self.update_status("Error saving recording")

    def update_status(self, message):
        self.status_bar.config(text=message)
        self.root.update_idletasks()

    def cleanup(self):
        # Release all video captures
        for track in self.tracks:
            if track is not None and 'cap' in track:
                track["cap"].release()

if __name__ == "__main__":
    root = tk.Tk()
    app = MultiCamSwitcher(root)
    
    # Clean up on exit
    def on_closing():
        app.cleanup()
        root.destroy()
    
    root.protocol("WM_DELETE_WINDOW", on_closing)
    root.mainloop()
